﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Core;
using HeuristicLab.Common;
using HeuristicLab.Hive.Contracts.Interfaces;
using HeuristicLab.Clients.Common;
using HeuristicLab.Hive.ExperimentManager;
using HeuristicLab.Hive.Contracts.BusinessObjects;
using HeuristicLab.PluginInfrastructure;
using HeuristicLab.Hive.Contracts.ResponseObjects;
using System.Threading;
using HeuristicLab.Random;
using System.Threading.Tasks;

namespace HeuristicLab.HiveEngine {
  /// <summary>
  /// Represents an engine that executes operations which can be executed in parallel on the hive
  /// </summary>
  [StorableClass]
  [Item("Hive Engine", "Engine for parallel execution on the hive. You need enable `Parallel` for at least one operator in your operator graph to have all childoperations parallelized. Also those childoperations must not have sideeffects on a higher scope.")]
  public class HiveEngine : Engine {
    private Semaphore maxConcurrentConnections = new Semaphore(4, 4); // avoid too many connections
    private Semaphore maxConcurrentSerializations = new Semaphore(1, 1); // allow ony ONE concurrent serialization, because the operations share the same ParentScopes and serializing the same objects concurrently causes problems
    private Semaphore maxSerializedJobsInMemory = new Semaphore(2, 2); // avoid memory problems
    private CancellationToken cancellationToken;

    [Storable]
    private IOperator currentOperator;

    [Storable]
    public string ResourceIds { get; set; }

    #region constructors and cloning
    public HiveEngine() {
      ResourceIds = "HEAL";
    }
    [StorableConstructor]
    protected HiveEngine(bool deserializing) : base(deserializing) { }
    protected HiveEngine(HiveEngine original, Cloner cloner)
      : base(original, cloner) {
      this.ResourceIds = original.ResourceIds;
      this.currentOperator = cloner.Clone(original.currentOperator);
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new HiveEngine(this, cloner);
    }
    #endregion

    protected override void Run(CancellationToken cancellationToken) {
      this.cancellationToken = cancellationToken;
      Run(ExecutionStack);
    }

    private void Run(object state) {
      Stack<IOperation> executionStack = (Stack<IOperation>)state;
      IOperation next;
      OperationCollection coll;
      IAtomicOperation operation;

      while (ExecutionStack.Count > 0) {
        cancellationToken.ThrowIfCancellationRequested();

        next = ExecutionStack.Pop();
        if (next is OperationCollection) {
          coll = (OperationCollection)next;
          if (coll.Parallel) {
            IDictionary<IOperation, OperationJob> jobs = new Dictionary<IOperation, OperationJob>();
            foreach (IOperation op in coll) {
              jobs.Add(op, new OperationJob(op));
            }

            ExecuteOnHive(jobs);

            foreach (var kvp in jobs) {
              if (kvp.Key is IAtomicOperation) {
                ReIntegrateScope(kvp.Value.Operation as IAtomicOperation, kvp.Key as IAtomicOperation);
              } else if (kvp.Key is OperationCollection) {
                // todo ??
              }
            }
          } else {
            for (int i = coll.Count - 1; i >= 0; i--)
              if (coll[i] != null) executionStack.Push(coll[i]);
          }
        } else if (next is IAtomicOperation) {
          operation = (IAtomicOperation)next;
          try {
            next = operation.Operator.Execute((IExecutionContext)operation, cancellationToken);
          }
          catch (Exception ex) {
            ExecutionStack.Push(operation);
            if (ex is OperationCanceledException) throw ex;
            else throw new OperatorExecutionException(operation.Operator, ex);
          }
          if (next != null) ExecutionStack.Push(next);

          if (operation.Operator.Breakpoint) {
            LogMessage(string.Format("Breakpoint: {0}", operation.Operator.Name != string.Empty ? operation.Operator.Name : operation.Operator.ItemName));
            Pause();
          }
        }
      }
    }

    private IRandom FindRandomParameter(IExecutionContext ec) {
      try {
        if (ec == null)
          return null;

        foreach (var p in ec.Parameters) {
          if (p.Name == "Random" && p is IValueParameter)
            return ((IValueParameter)p).Value as IRandom;
        }
        return FindRandomParameter(ec.Parent);
      }
      catch { return null; }
    }

    private static void ReIntegrateScope(IAtomicOperation source, IAtomicOperation target) {
      ExchangeScope(source.Scope, target.Scope);
    }

    private static void ExchangeScope(IScope source, IScope target) {
      target.Variables.Clear();
      target.Variables.AddRange(source.Variables);
      target.SubScopes.Clear();
      target.SubScopes.AddRange(source.SubScopes);
      // TODO: validate if parent scopes match - otherwise source is invalid
    }

    /// <summary>
    /// This method blocks until all jobs are finished
    /// </summary>
    /// <param name="jobDict"></param>
    private void ExecuteOnHive(IDictionary<IOperation, OperationJob> jobDict) {
      LogMessage(string.Format("Executing {0} operations on the hive.", jobDict.Count));

      object locker = new object();
      try {
        IDictionary<Guid, IOperation> jobIds = new Dictionary<Guid, IOperation>();
        List<Guid> remainingJobIds = new List<Guid>();
        IDictionary<Guid, int> jobNumbers = new Dictionary<Guid, int>(); // for better readability of log
        JobResultList results;
        var pluginsNeeded = ApplicationManager.Manager.Plugins.Select(x => new HivePluginInfoDto { Name = x.Name, Version = x.Version }).ToList();
        int finishedCount = 0;

        // create upload-tasks
        var uploadTasks = new List<Task<JobDto>>();
        foreach (var kvp in jobDict) {
          // shuffle random variable to avoid the same random sequence in each operation; todo: does not yet work (it cannot find the random variable)
          IRandom random = FindRandomParameter(kvp.Key as IExecutionContext);
          if (random != null)
            random.Reset(random.Next());

          uploadTasks.Add(Task.Factory.StartNew<JobDto>((keyValuePairObj) => {
            return UploadJob(pluginsNeeded, keyValuePairObj);
          }, kvp));
        }

        Task processUploadedJobsTask = Task.Factory.StartNew(() => {
          // process finished upload-tasks
          int uploadTasksCount = uploadTasks.Count;
          for (int i = 0; i < uploadTasksCount; i++) {
            var uploadTasksArray = uploadTasks.ToArray();
            var task = uploadTasksArray[Task.WaitAny(uploadTasksArray)];
            if (task.Status == TaskStatus.Faulted) {
              LogException(task.Exception);
              throw task.Exception;
            }

            IOperation key = ((KeyValuePair<IOperation, OperationJob>)task.AsyncState).Key;
            JobDto jobDto = task.Result;

            jobIds.Add(jobDto.Id, key);
            remainingJobIds.Add(jobDto.Id);
            jobNumbers.Add(jobDto.Id, remainingJobIds.Count);

            LogMessage(string.Format("Submitted job #{0}", jobNumbers[jobDto.Id], jobDto.Id));
            uploadTasks.Remove(task);
          }
        });

        // poll job-statuses and create tasks for those which are finished
        var downloadTasks = new List<Task<OperationJob>>();
        var executionTimes = new List<TimeSpan>();
        while (processUploadedJobsTask.Status != TaskStatus.RanToCompletion || remainingJobIds.Count > 0) {
          Thread.Sleep(10000);
          using (Disposable<IClientFacade> service = ServiceLocator.Instance.ClientFacadePool.GetService()) {
            results = service.Obj.GetJobResults(remainingJobIds).Obj;
          }
          var jobsFinished = results.Where(j => j.State == JobState.Finished || j.State == JobState.Failed || j.State == JobState.Aborted);
          finishedCount += jobsFinished.Count();
          var totalExecutionTime = TimeSpan.FromMilliseconds(results.Select(j => j.ExecutionTime).Union(executionTimes).Select(e => e.TotalMilliseconds).Sum());
          LogMessage(string.Format("Results polled. Jobs finished: {0}/{1}, TotalExecutionTime: {2}", finishedCount, jobDict.Count, totalExecutionTime));
          foreach (var result in jobsFinished) {
            if (result.State == JobState.Finished) {
              downloadTasks.Add(Task.Factory.StartNew<OperationJob>((jobIdObj) => {
                return DownloadJob(jobNumbers, jobIdObj);
              }, result.Id));
            } else if (result.State == JobState.Aborted) {
              LogMessage(string.Format("Job #{0} aborted (id: {1})", jobNumbers[result.Id], result.Id));
            } else if (result.State == JobState.Failed) {
              LogMessage(string.Format("Job {0} failed (id: {1}): {2}", jobNumbers[result.Id], result.Id, result.Exception));
            }
            remainingJobIds.Remove(result.Id);
            executionTimes.Add(result.ExecutionTime);
          }
        }

        // process finished download-tasks
        int downloadTasksCount = downloadTasks.Count;
        for (int i = 0; i < downloadTasksCount; i++) {
          var downloadTasksArray = downloadTasks.ToArray();
          var task = downloadTasksArray[Task.WaitAny(downloadTasksArray)];
          var jobId = (Guid)task.AsyncState;
          if (task.Status == TaskStatus.Faulted) {
            LogException(task.Exception);
            throw task.Exception;
          }
          jobDict[jobIds[(Guid)task.AsyncState]] = task.Result;
          downloadTasks.Remove(task);
        }

        LogMessage(string.Format("All jobs finished (TotalExecutionTime: {0}). Deleting jobs on hive.", TimeSpan.FromMilliseconds(executionTimes.Select(e => e.TotalMilliseconds).Sum())));
        // delete jobs
        using (Disposable<IClientFacade> service = ServiceLocator.Instance.ClientFacadePool.GetService()) {
          foreach (Guid jobId in jobIds.Keys) {
            service.Obj.DeleteJob(jobId);
          }
        }

        LogMessage(string.Format("Operations on the hive finished.", jobDict.Count));
      }
      catch (Exception e) {
        LogException(e);
        throw e;
      }
    }

    private JobDto UploadJob(List<HivePluginInfoDto> pluginsNeeded, object keyValuePairObj) {
      var keyValuePair = (KeyValuePair<IOperation, OperationJob>)keyValuePairObj;
      var groups = ResourceIds.Split(';');
      maxSerializedJobsInMemory.WaitOne();
      SerializedJob serializedJob = new SerializedJob();
      maxConcurrentSerializations.WaitOne();
      serializedJob.SerializedJobData = SerializedJob.Serialize(keyValuePair.Value);
      maxConcurrentSerializations.Release();
      serializedJob.JobInfo = new JobDto();
      serializedJob.JobInfo.State = JobState.Offline;
      serializedJob.JobInfo.CoresNeeded = 1;
      serializedJob.JobInfo.PluginsNeeded = pluginsNeeded;
      ResponseObject<JobDto> response;
      maxConcurrentConnections.WaitOne();
      using (Disposable<IClientFacade> service = ServiceLocator.Instance.StreamedClientFacadePool.GetService()) {
        response = service.Obj.AddJobWithGroupStrings(serializedJob, groups);
        serializedJob = null;
        maxSerializedJobsInMemory.Release();
      }
      maxConcurrentConnections.Release();
      return response.Obj;
    }

    private OperationJob DownloadJob(IDictionary<Guid, int> jobNumbers, object jobIdObj) {
      Guid jobId = (Guid)jobIdObj;
      SerializedJob serializedJob;
      maxConcurrentConnections.WaitOne();
      maxSerializedJobsInMemory.WaitOne();
      using (Disposable<IClientFacade> service = ServiceLocator.Instance.StreamedClientFacadePool.GetService()) {
        serializedJob = service.Obj.GetLastSerializedResult(jobId).Obj;
      }
      maxConcurrentConnections.Release();
      OperationJob operationJob = SerializedJob.Deserialize<OperationJob>(serializedJob.SerializedJobData);
      serializedJob = null;
      maxSerializedJobsInMemory.Release();
      LogMessage(string.Format("Downloaded job #{0}", jobNumbers[jobId], jobId));
      return operationJob;
    }

    /// <summary>
    /// Threadsafe message logging
    /// </summary>
    private void LogMessage(string message) {
      lock (Log) {
        Log.LogMessage(message);
      }
    }

    /// <summary>
    /// Threadsafe exception logging
    /// </summary>
    private void LogException(Exception exception) {
      lock (Log) {
        Log.LogException(exception);
      }
    }
  }
}
